import unittest

import preprocessing
import eval


class TestPreprocssing(unittest.TestCase):

    def testfun(self):
        langs = ['ab', 'ar', 'as', 'br', 'ca', 'cnh', 'cs', 'cv', 'cy', 'de', 'dv', 'el', 'en', 'eo', 'es', 'et', 'eu',
                 'fa',
                 'fi', 'fr', 'fy-NL', 'ga-IE', 'hi', 'hsb', 'hu', 'ia', 'id', 'it', 'ja', 'ka', 'kab', 'ky', 'lg', 'lt',
                 'lv',
                 'mn', 'mt', 'nl', 'or', 'pa-IN', 'pl', 'pt', 'rm-sursilv', 'rm-vallader', 'ro', 'ru', 'rw', 'sah',
                 'sl',
                 'sv-SE', 'ta', 'th', 'tr', 'tt', 'uk', 'vi', 'vot', 'zh-CN', 'zh-HK', 'zh-TW']

        count = 0
        total = 0
        for lang in langs:
            total += 1
            try:
                batch = {}
                batch['sentence'] = "hello 你好嗎 図書館にいた事がバレた , , ? . !;: “%”�()«»½…"
                print("fun name", 'fun_' + lang.replace("-", "_"))
                preprocessing_sentence = getattr(preprocessing, 'fun_' + lang.replace("-", "_"))
                print(preprocessing_sentence(batch))
                count += 1
            except BaseException as e:
                print(e)
                pass
        print(count, total)

    def test_regex(self):
        langs = ['ab', 'ar', 'as', 'br', 'ca', 'cnh', 'cs', 'cv', 'cy', 'de', 'dv', 'el', 'en', 'eo', 'es', 'et', 'eu',
                 'fa',
                 'fi', 'fr', 'fy-NL', 'ga-IE', 'hi', 'hsb', 'hu', 'ia', 'id', 'it', 'ja', 'ka', 'kab', 'ky', 'lg', 'lt',
                 'lv',
                 'mn', 'mt', 'nl', 'or', 'pa-IN', 'pl', 'pt', 'rm-sursilv', 'rm-vallader', 'ro', 'ru', 'rw', 'sah',
                 'sl',
                 'sv-SE', 'ta', 'th', 'tr', 'tt', 'uk', 'vi', 'vot', 'zh-CN', 'zh-HK', 'zh-TW']

        count = 0
        total = 0
        re_list = []
        for lang in langs:
            total += 1
            try:
                batch = {}
                batch['sentence'] = "hello 你好嗎 図書館にいた事がバレた , , ? . !;: “%”�()«»½…"
                print("fun name", 'fun_' + lang.replace("-", "_"))
                preprocessing_sentence = getattr(preprocessing, 'fun_' + lang.replace("-", "_"))
                print(preprocessing_sentence(batch))
                if preprocessing_sentence.chars_to_ignore_regex not in re_list:
                    re_list.append(preprocessing_sentence.chars_to_ignore_regex)
                count += 1
            except BaseException as e:
                print(e)
                pass
        generic_re = '|'.join(re_list)
        print(generic_re)
        print(count, total)


class TestEval(unittest.TestCase):
    def testCER(self):
        targets = ['HuggingFace is great!', 'Love Transformers!', 'Let\'s wav2vec!']
        preds = ['HuggingFace is awesome!', 'Transformers is powerful.', 'Let\'s finetune wav2vec!']
        print("CER: {:2f}".format(100 * eval.cer(targets, preds)))
        print("chunk size = None, CER: {:2f}".format(100 * eval.chunked_cer(targets, preds, chunk_size=None)))
        print("chunk size = 2, CER: {:2f}".format(100 * eval.chunked_cer(targets, preds, chunk_size=2)))
        print("chunk size = 3, CER: {:2f}".format(100 * eval.chunked_cer(targets, preds, chunk_size=3)))
        print("chunk size = 5, CER: {:2f}".format(100 * eval.chunked_cer(targets, preds, chunk_size=5)))
        print("chunk size = 7, CER: {:2f}".format(100 * eval.chunked_cer(targets, preds, chunk_size=6)))
        print("chunk size = 100, CER: {:2f}".format(100 * eval.chunked_cer(targets, preds, chunk_size=100)))

    def testBoth(self):
        targets = ["hello world", "你好啊"]
        preds = ["hello duck", "您好嗎"]
        print("WER: {:2f}".format(100 * eval.wer(targets, preds)))
        print("chunk size = None, WER: {:2f}".format(100 * eval.chunked_wer(targets, preds, chunk_size=None)))
        print("chunk size = None, CER: {:2f}".format(100 * eval.chunked_cer(targets, preds, chunk_size=None)))



import jiwer
import editdistance as ed


def cer(groundtruth, hypothesis):
    err = 0
    tot = 0
    for p, t in zip(hypothesis, groundtruth):
        err += float(ed.eval(p.lower(), t.lower()))
        tot += len(t)
    return err / tot


def wer(groundtruth, hypothesis):
    err = 0
    tot = 0
    for p, t in zip(hypothesis, groundtruth):
        p = p.lower().split(' ')
        t = t.lower().split(' ')
        err += float(ed.eval(p, t))
        tot += len(t)
    return err / tot


def chunked_cer(targets, predictions, chunk_size=None):
    _predictions = [char for seq in predictions for char in list(seq)]
    _targets = [char for seq in targets for char in list(seq)]
    if chunk_size is None: return jiwer.wer(_targets, _predictions)
    start = 0
    end = chunk_size
    H, S, D, I = 0, 0, 0, 0
    while start < len(targets):
        _predictions = [char for seq in predictions[start:end] for char in list(seq)]
        _targets = [char for seq in targets[start:end] for char in list(seq)]
        chunk_metrics = jiwer.compute_measures(_targets, _predictions)
        H = H + chunk_metrics["hits"]
        S = S + chunk_metrics["substitutions"]
        D = D + chunk_metrics["deletions"]
        I = I + chunk_metrics["insertions"]
        start += chunk_size
        end += chunk_size
    return float(S + D + I) / float(H + S + D)


def chunked_wer(targets, predictions, chunk_size=None):
    if chunk_size is None: return jiwer.wer(targets, predictions)
    start = 0
    end = chunk_size
    H, S, D, I = 0, 0, 0, 0
    while start < len(targets):
        chunk_metrics = jiwer.compute_measures(targets[start:end], predictions[start:end])
        H = H + chunk_metrics["hits"]
        S = S + chunk_metrics["substitutions"]
        D = D + chunk_metrics["deletions"]
        I = I + chunk_metrics["insertions"]
        start += chunk_size
        end += chunk_size
    return float(S + D + I) / float(H + S + D)



from itertools import groupby

import joblib
import torch
from transformers import Wav2Vec2FeatureExtractor, HubertModel
import soundfile as sf


class HubertCode(object):
    def __init__(self, hubert_model, km_path, km_layer, return_diff=False):
        self.processor = Wav2Vec2FeatureExtractor.from_pretrained(hubert_model)
        self.model = HubertModel.from_pretrained(hubert_model)
        self.model.eval()
        self.km_model = joblib.load(km_path)
        self.km_layer = km_layer
        self.return_diff = return_diff
        self.C_np = self.km_model.cluster_centers_.transpose()
        self.Cnorm_np = (self.C_np ** 2).sum(0, keepdims=True)

        self.C = torch.from_numpy(self.C_np)
        self.Cnorm = torch.from_numpy(self.Cnorm_np)
        if torch.cuda.is_available():
            self.C = self.C.cuda()
            self.Cnorm = self.Cnorm.cuda()
            self.model = self.model.cuda()

    def __call__(self, filepath, sampling_rate=None, merge=True):
        with torch.no_grad():
            speech, sr = sf.read(filepath)
            input_values = self.processor(speech, return_tensors="pt", sampling_rate=sr).input_values
            if torch.cuda.is_available():
                input_values = input_values.cuda()
            hidden_states = self.model(input_values, output_hidden_states=True).hidden_states
            x = hidden_states[self.km_layer].squeeze()
            dist = torch.sqrt(
                x.pow(2).sum(1, keepdim=True)
                - 2 * torch.matmul(x, self.C)
                + self.Cnorm
            )
            min_dist = dist.detach().min(dim=1)
            if merge:
                unitcode = [k for k, _ in groupby(min_dist.indices.cpu().numpy())]
            else:
                unitcode = min_dist.indices.cpu().numpy()
            if self.return_diff:
                return unitcode, x.cpu() - torch.index_select(
                    torch.tensor(self.C_np.transpose()).cpu(), 0, min_dist.indices.cpu())
            else:
                return unitcode



from setuptools import setup, find_packages

with open('requirements.txt') as f:
    required = f.read().splitlines()

setup(
    name='asrp',
    version='0.0.11',
    description='',
    url='https://github.com/voidful/asrp',
    author='Voidful',
    author_email='voidful.stack@gmail.com',
    long_description=open("README.md", encoding="utf8").read(),
    long_description_content_type="text/markdown",
    setup_requires=['setuptools-git'],
    classifiers=[
        'Development Status :: 4 - Beta',
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python"
    ],
    license="Apache",
    keywords='asr',
    packages=find_packages(),
    install_requires=required,
    zip_safe=False,
)
